//	PWM 2 Stepper converter - for megasquirt users


//	Michieletti Enrico
//	v1.0 - Initial release
//	v1.1 - instead of interrupt the flow every 0,1ms, i'll interrupt it only on rb3 pin change (CCP Compare mode)
//		 - for stepper moviment, i'll done an interrupt when the pr2 register match the right step time
//		 - various improvements
//		 - used tmr1 for pwm acquisition. uS accuracy (instead of 100uS), min freq = 1/(65536uS) = 15hz! Max freq = ? probably hundreds khz region.
//		 - 15 minutes IAC always on, after this, moving only
//		 - Moved pwm input to rb0 to rb3 (for use the 16 bit 'capture' mode)
//		 - Moved also IACDis, Phase1 and Phase2
//		 - There is still an approssimation on the tmr2: 96uS instead of 100... So, for 3.0ms we get 120uS of error...
//	v1.2 - Added line load for a/c. When this line goes from low to high, the stepper retract of n steps for x time

#include	<pic.h>

__CONFIG(WDTDIS & PWRTEN & XT & MCLRDIS & BOREN & LVPDIS);		//Fuses configuration

__EEPROM_DATA(170, 30, 0, 1, 10, 100, 30, 1);
//1  is the max step that the iac use to go from fully open to fully closed (from 0 to 255 - useless over 255! waste of ram and resources)
//2	 is the step time (in 0,1 ms) - from 0 to 255(25,5ms)
//3, if set to 0 move only the motor, 1 leave always power to the motor (always on), 2 leaves for 15min the power at the motor, then moving only
//4, is the change that must have the pwm (in %) before the stepper will move - from 0 to 100 (ex: if we are at 50% dc, with 5% thr, the stepper moves if signal falls less than 45 or rise more than 55)
//5, when you maintain pressed a button for up/down, this is the delay from one step to another (must be higher than step time, if not is used the step time value) - from 0 to 255(255ms)
//6, Number of useful steps (0 - 255)
//7, number of steps that retract for n time when there is a signal on the load line (0 for disable)
//8, n time (in seconds) of increased idle

//Declare the function
int LinearInterpolate(char X1,char Y1,char X2,char Y2,char XX);

//X1 is a char (DC 0-100)
//Y1 is a int (step)\
//X2 is a char (DC 0-100)
//Y2 is a int (step)
//XX is a char (DC 0-100)

//inputs
#define		Down 	RB5		//button for extend the pintle (decrease idle rpm)	|Pressed together change	
#define		Up 		RB4		//button for retract the pintle (increase idle rpm)	|the idle mode (auto/manual)
#define		PWMIN 	RB3		//input for pwn acquire
#define		Load 	RA5		//input for load

//outputs
#define		ModeLed	RB6		//led for mode: off-automatic idle (default) / on-manual idle
#define		IACDis 	RB0		//IAC enable (used for always-on/moving only)
#define		Phase1 	RB1		//|Phases for stepper moviment
#define		Phase2 	RB2		//|

static volatile unsigned char MaxSteps;				//numbers of steps to extend for first run
static volatile unsigned char StepTime;				//step time (in 0,1ms)
static volatile unsigned char AlwaysOn;				//0 = moving only / 1=always on / 2=15min always on
static volatile unsigned char PWMthr;				//pwm threshold before doing a step change (%)
static volatile unsigned char CurrDC = 0;			//current duty cycle (%)
static volatile unsigned char CurrStep;				//Current step
static volatile int MoveSteps = 0;					// > 0 move n steps to decrease idle; < 0 move n steps do increase idle - must be signed integer!
static volatile unsigned char BDelay;				//button delay (in ms)
static volatile unsigned char TMPBDelay = 1;		//temp variable for count the button delay (in ms)
static volatile unsigned char MaxUsableSteps;		//Maximum usable steps (The iac already full flow without be all open)
static volatile unsigned long TMP15min = 0;			//temp variable for count 900.000 ticks (in 0,1 ms section) needed for the 15min iac
static volatile unsigned int T = 0;					//Period of the frequency
static volatile unsigned int Ton = 0;				//Variable for The current High period
static volatile unsigned int TC = 0;				//Temp. Period of the frequency - used for calculations
static volatile unsigned long TonC = 0;				//Temp. Variable for The current High period - used for calculation
static volatile unsigned char PrevDC = 0;			//Prevoius DC, for threshold use
static volatile bit UnderLoad = 0;					//Variable for know when we are under load
static volatile unsigned char LoadSteps = 0;		//Steps to retract (for increase idle) when under load
static volatile unsigned char LoadSecs = 0;			//Seconds for maintain the idle increased when under load (max 10 secs)
static volatile unsigned int TMPLoadSecs = 0;

static volatile bit Mode = 0;						//0 = Auto; 1 = Manual
static volatile bit PrevUp = 0;						//Previous button state, for not entering 2 times in the same routine
static volatile bit PrevDown = 0;					//in the same routine
static volatile bit PrevLoad = 0;					//in the same routine

static volatile bit	EDGE @ (unsigned)&CCP1CON*8+0;	//if 1, is in rising edge

interrupt isr() {
	if (CCP1IE && CCP1IF) {				//change of PWM IN pin
		if (EDGE == 1) {
			//interrupt on the rising edge
			//Here i sample the total time
			T = CCPR1H * 256 + CCPR1L;
			TMR1H = 0;		//|Reset the timer
			TMR1L = 0;		//|
		} else {
			//interrupt on the falling edge
			//here i sample the Ton
			Ton = CCPR1H * 256 + CCPR1L;	//Save the On time
		}
		CCP1IF=0;
		EDGE = !EDGE;
	}
	if (INTE && INTF) {		//If there is an interrupt on the rb0 pin
		//no more here!
	}
	if (TMR1IE && TMR1IF) {	//the TMR1 has gone in overflow, dc at 0% or 100%
		if (EDGE == 1) {
			//stuck waiting rising edge, assuming dc = 0%
			Ton = 0;
			T = 65535;
		} else {
			//stuck waiting falling edge, assuming dc = 100%
			Ton = 65535;
			T = 65535;
		}
		// The tmr1 counter reset itself (rollover)		
		TMR1IF = 0;
	}

	if (T0IE && T0IF) {		//timer0 interrupt (1ms)
		TMR0 =13;	//is used for the delay of the button pressed and load line control, start at 13 for 1000uS (else 1024uS)
		if (TMPBDelay == BDelay) {
			//if I press both buttons, i change mode (Auto/Manual) (Enter only one time!)
			
			if ((Up == 0 && Down == 0) && ((Up != PrevUp) || (Down != PrevDown))) {
				Mode = !Mode;
				ModeLed = Mode;
				if (Mode == 0) {
					//if we return in auto, i'll force it to recalculate the moviment
					PrevDC = 100 - CurrDC;		//i can't done this... WHY???
				} else {
					//if we return in manual, i'll reset the UnderLoad flag
					UnderLoad = 0;
				}
			} else {
				if (Mode == 1) {		//If we are in manual mode
					if (Up == 0) MoveSteps += 1;		//Retract the pintle - More idle speed
					if (Down == 0) MoveSteps -= 1;		//Extend the pintle - Less idle speed
				}
			}
			PrevUp = Up;
			PrevDown = Down;			
			
			TMPBDelay = 1;	
		} else {
			TMPBDelay ++;
		}


		//Here i place the section for the load line

		//Checking Input
		if ((Load == 1) && (Load != PrevLoad) && (Mode == 0) && (LoadSteps > 0)) {
			//enter here only one time whe the lines goes up, and if we are in auto mode

			//charge the timer
			TMPLoadSecs = LoadSecs * 1000;
			
			//if we are already under load...
			if (UnderLoad == 1) {
				//we are already under load, nothing to do...
			} else {
				//Set the UnderLoad flag
				UnderLoad = 1;
				//move the stepper...
				MoveSteps = MoveSteps + LoadSteps;
			}

		}
		PrevLoad = Load;
		
		//Here is the count
		if (UnderLoad == 1) {
			//If we are under load
			if (TMPLoadSecs == 0) {
				//if the time is elapsed i'll reset the UnderLoad Flag
				//this reengage the normal DC calculation
				UnderLoad = 0;
				//and force to recalculate the dc
				PrevDC = 100 - CurrDC;		//i can't done this... WHY???				
			} else {
				TMPLoadSecs --;
			}
		}


		//here i place the routine for the 15 minute iac.
		//1ms section, so: 15min = 900sec = 900.000ms!!! 
		if (AlwaysOn == 2) {

			if (TMP15min == 900000) {
				AlwaysOn = 0;			//switch to moving only
				TMP15min = 0;	
			} else {
				TMP15min ++;
			}
		}
		

		T0IF = 0;
	}

	if (TMR2IE && TMR2IF) {	//if the TMR2 has reached the right count (96us * steptime)
		//the moviment of the stepper is done here...
		if (MoveSteps > 0) {
			if (CurrStep == MaxUsableSteps) {
				//prevent any movement over the allowed range
				MoveSteps = 0;
			} else {
				//move the stepper retracting the pintle
				//calculating the next step basing on the prevoius
				if (Phase1 == 0 && Phase2 == 0)	{
					Phase1 = 1;
					Phase2 = 0;
				} else if (Phase1 == 0 && Phase2 == 1) {
					Phase1 = 0;
					Phase2 = 0;
				} else if (Phase1 == 1 && Phase2 == 0) {
					Phase1 = 1;
					Phase2 = 1;
				} else if (Phase1 == 1 && Phase2 == 1) {
					Phase1 = 0;
					Phase2 = 1;
				}

				//enable the stepper chip
				IACDis = 0;
				//Increase CurrStep
				CurrStep += 1;
				//Decrease MoveStep
				MoveSteps -= 1;
			} 
		} else if (MoveSteps < 0) {
			if (CurrStep == 0) {
				//prevent any movement over the allowed range
				MoveSteps = 0;
			} else {
				//move the stepper extending the pintle
				//calculating the next step basing on the prevoius
				if (Phase1 == 0 && Phase2 == 0)	{
					Phase1 = 0;
					Phase2 = 1;
				} else if (Phase1 == 0 && Phase2 == 1) {
					Phase1 = 1;
					Phase2 = 1;
				} else if (Phase1 == 1 && Phase2 == 0) {
					Phase1 = 0;
					Phase2 = 0;
				} else if (Phase1 == 1 && Phase2 == 1) {
					Phase1 = 1;
					Phase2 = 0;
				}

				//enable the stepper chip
				IACDis = 0;
				//Decrease CurrStep
				CurrStep -= 1;
				//Increase MoveStep
				MoveSteps += 1;
			} 
		} else {
			//Taking care of the always-on. if alwayson = 0 then shut off, else leave on
			if (AlwaysOn == 0) {
				IACDis = 1;
			} else {
				IACDis = 0;
			}
		}
	TMR2IF=0;
	}
}

main() {
	TRISA = 0B00000000;		//porta, all outputs (not used)
	TRISB = 0B00111000;		//portb
	PORTA = 0;				//Resets all the outputs
	PORTB = 0;
	IACDis = 1;				//disable the Iac
	ModeLed = 0;

	//load variables from eeprom
	MaxSteps = EEPROM_READ(00) ;
	StepTime = EEPROM_READ(01);
	AlwaysOn = EEPROM_READ(02);
	PWMthr = EEPROM_READ(03);
	BDelay = EEPROM_READ(04);
	MaxUsableSteps = EEPROM_READ(05);
	LoadSteps = EEPROM_READ(06);
	LoadSecs = EEPROM_READ(07);

	
	OPTION = 0B00000001;	//7 Enable portb pullup = 0!
							//6 not used - rb0 int edge
							//5 tmr0 from clock
							//4 
							//3 prescaler to tmr0 (for control the buttons)
							//2 |prescaler to 4 (256 * 4 = 1024us)
							//1 |the sw control inputs every 1ms (* bdelay) the buttons 
							//0 |
	INTCON = 0B11100000;	//7 disable global interrupts 
							//6 enable peripherals interrupts
							//5 Interrupt on TMR0 - only when the tmr0 is not reset (dc 0% or dc 100%)
							//4 disable Interrupt on RB0
							//3 disable interrupt on portb change
							//2 Reset tmr0 Interrupt flag
							//1 Reset RB0 Interrupt flag 
							//0 Reset RB4-RB7 interrupt flag
	PIE1 = 0B00000111;		//7 Interrupt on eeprom flash complete disabled
							//6 Comparator interrupt disabled
							//5 disable usart receive interrupt
							//4 disable usart transmit interrupt
							//3 not used
							//2 Enable CCP interrupt
							//1 Enable the tmr2 match interrupt
							//0 tmr1 overflow interrupt
	T2CON = 0B00101111;		//7 not used
							//6 | postscaler 1:5 (give us 96uS interrupt)
							//5 |
							//4 |
							//3 |
							//2 tmr2 enabled
							//1 | Prescaler 1:16 (16uS)
							//0 |
	CCP1CON = 0B00000101;	//7 not used
							//6 not used
							//5 | not used in compare
							//4 |
							//3 | Configure for capture mode, every rising edge
							//2 | (the edge is changed in code)
							//1 |
							//0 |
	T1CON = 0B00000001;		//7 not used
							//6 not used
							//5 | Prescaler 1:1 (1uS accuracy)
							//4 |
							//3 TRM1 oscillator disabled
							//2 Doesn't sync on ext source
							//1 Internal clock
							//0 TMR1 enabled


	//if i set a pause for reading the buttons smaller than the step time, 
	if ((BDelay * 10) < StepTime) BDelay = (StepTime / 10) + 1;
	if (MaxUsableSteps > MaxSteps) MaxUsableSteps = MaxSteps;
	if (LoadSecs > 10) LoadSecs = 10;

	PR2 = StepTime - 1;	//prescaler 1:16 * 6 (0 to 5 postscaler) * StepTime = 96uS * StepTime

	//Resets the timers
	TMR0 = 13;	//is used for the delay of the button pressed, start at 6 for 1000uS (else 1024uS)
	TMR1H = 0;	//|is used for accurate measure the pwm (16 bit!)
	TMR1L = 0;	//|
	TMR2 = 0;	//is used for the stepper moviment
	
	GIE = 1;	//enable global interrupts

	//Move (maxsteps) forward for decrease the idle (and let the stepper to goes on the 0 position)
	CurrStep = MaxSteps;			//Preventing negative numbers
	MoveSteps = - MaxSteps;	


init:
	//the first time wait until the stepper goes to full extend
	if (MoveSteps != 0) goto init;
	//here i set the 0 position
	CurrStep = 0;

	
start:
	//if i'm no moving the stepper and i'm in auto mode
	if (MoveSteps == 0 && Mode == 0 && UnderLoad == 0) {
		//the duty cycle calculation is done here, too slow for done it in the isr

		TonC = Ton;
		TC = T;
		if (TonC != 0 && TC != 0) {

			CurrDC = (TonC * 100) / TC ;
			TonC = 0;
			TC = 0;
		}

		//probably i need something that prevent the threshold to ostacolate
		// the full open or full close of the pintle (when we are sure that dc = 100 or = 0)... 
		if ((CurrDC == 100) || (CurrDC == 0)) goto calc;

		//Threshold section...
		if (CurrDC < PrevDC) goto inv;
		if ((CurrDC - PrevDC) > PWMthr) goto calc;
inv:
		if ((PrevDC - CurrDC) > PWMthr) goto calc;
		goto start;		

calc:
			//i calculate with interpulation the relative stepper moviment (realted to DC%)
			//DC = 0% -> Stepper all extended (0 = initial position)
			//DC = 100% -> Stepper all retracted (maxsteps)

			PrevDC = CurrDC;
			MoveSteps = LinearInterpolate(0,0,100,MaxUsableSteps,CurrDC) - CurrStep;
	}
	goto start;
}

//Interpolation function..
int LinearInterpolate(char X1,char Y1,char X2,char Y2,char XX) {
	return ((XX - X1) * (Y2 - Y1) + Y1 * (X2 - X1)) / (X2 - X1);
}
